@@ -5,7 +5,7 @@
  * $Author: kou $
  * $Date: 2008-09-26 13:52:08 $
  *
- * Copyright 2005-2018 Kouhei Sutou <kou@cozmixng.org>
+ * Copyright 2005-2022 Sutou Kouhei <kou@cozmixng.org>
  * Copyright 2005 Øyvind Kolås <pippin@freedesktop.org>
  * Copyright 2004-2005 MenTaLguY <mental@rydia.com>
  *
@@ -40,6 +40,23 @@ cr_check_status (cairo_t *context)
   rb_cairo_check_status (cairo_status (context));
 }
 
+static void
+cr_context_free (void *ptr)
+{
+  cairo_destroy ((cairo_t *) ptr);
+}
+
+static const rb_data_type_t cr_context_type = {
+  "Cairo::Context",
+  {
+    NULL,
+    cr_context_free,
+  },
+  NULL,
+  NULL,
+  RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
 /* Functions for manipulating state objects */
 cairo_t *
 rb_cairo_context_from_ruby_object (VALUE obj)
@@ -49,7 +66,7 @@ rb_cairo_context_from_ruby_object (VALUE obj)
     {
       rb_raise (rb_eTypeError, "not a cairo graphics context");
     }
-  Data_Get_Struct (obj, cairo_t, context);
+  TypedData_Get_Struct (obj, cairo_t, &cr_context_type, context);
   if (!context)
     rb_cairo_check_status (CAIRO_STATUS_NULL_POINTER);
   return context;
@@ -67,22 +84,13 @@ cr_object_holder_free (void *ptr)
   rb_cairo__object_holder_free (rb_cCairo_Context, ptr);
 }
 
-static void
-cr_context_free (void *ptr)
-{
-  if (ptr)
-    {
-      cairo_destroy ((cairo_t *) ptr);
-    }
-}
-
 VALUE
 rb_cairo_context_to_ruby_object (cairo_t *cr)
 {
   if (cr)
     {
       cairo_reference (cr);
-      return Data_Wrap_Struct (rb_cCairo_Context, NULL, cr_context_free, cr);
+      return TypedData_Wrap_Struct (rb_cCairo_Context, &cr_context_type, cr);
     }
   else
     {
@@ -93,7 +101,7 @@ rb_cairo_context_to_ruby_object (cairo_t *cr)
 static VALUE
 cr_allocate (VALUE klass)
 {
-  return Data_Wrap_Struct (klass, NULL, cr_context_free, NULL);
+  return TypedData_Wrap_Struct (klass, &cr_context_type, NULL);
 }
 
 static void
@@ -118,20 +126,20 @@ cr_destroy (VALUE self)
   cr_set_user_data (cr, &cr_object_holder_key, NULL, NULL);
   cairo_destroy (cr);
 
-  DATA_PTR (self) = NULL;
+  RTYPEDDATA_DATA (self) = NULL;
   return Qnil;
 }
 
 static VALUE
 cr_destroyed (VALUE self)
 {
-  return CBOOL2RVAL (DATA_PTR (self) == NULL);
+  return CBOOL2RVAL (RTYPEDDATA_DATA (self) == NULL);
 }
 
 static VALUE
 cr_destroy_with_destroy_check (VALUE self)
 {
-  if (DATA_PTR (self))
+  if (RTYPEDDATA_DATA (self))
     cr_destroy (self);
   return Qnil;
 }
@@ -194,7 +202,7 @@ cr_s_wrap (VALUE self, VALUE pointer)
 
   rb_cr = rb_obj_alloc (self);
   cairo_reference (cr);
-  DATA_PTR (rb_cr) = cr;
+  RTYPEDDATA_DATA (rb_cr) = cr;
   rb_ivar_set (rb_cr, cr_id_surface, Qnil);
 
   if (rb_block_given_p ())
@@ -224,7 +232,7 @@ cr_initialize (VALUE self, VALUE target)
                       &cr_object_holder_key,
                       cr_object_holder_new (self),
                       cr_object_holder_free);
-  DATA_PTR (self) = cr;
+  RTYPEDDATA_DATA (self) = cr;
   if (rb_block_given_p ())
     result = rb_ensure (rb_yield, self, cr_destroy_with_destroy_check, self);
   return result;
@@ -240,6 +248,12 @@ cr_to_ptr (VALUE self)
                      1, PTR2NUM (_SELF));
 }
 
+static VALUE
+cr_raw_address (VALUE self)
+{
+  return PTR2NUM (_SELF);
+}
+
 
 static VALUE
 cr_restore (VALUE self)
@@ -351,7 +365,7 @@ cr_set_source_rgb (int argc, VALUE *argv, VALUE self)
   if (n == 1 && rb_cairo__is_kind_of (red, rb_cArray))
     {
       VALUE ary = red;
-      n = RARRAY_LEN (ary);
+      n = (int) RARRAY_LEN (ary);
       red = rb_ary_entry (ary, 0);
       green = rb_ary_entry (ary, 1);
       blue = rb_ary_entry (ary, 2);
@@ -388,7 +402,7 @@ cr_set_source_rgba (int argc, VALUE *argv, VALUE self)
   if (n == 1 && rb_cairo__is_kind_of (red, rb_cArray))
     {
       VALUE ary = red;
-      n = RARRAY_LEN (ary);
+      n = (int) RARRAY_LEN (ary);
       red = rb_ary_entry (ary, 0);
       green = rb_ary_entry (ary, 1);
       blue = rb_ary_entry (ary, 2);
@@ -517,6 +531,16 @@ cr_set_line_width (VALUE self, VALUE width)
   return self;
 }
 
+#if CAIRO_CHECK_VERSION(1, 17, 6)
+static VALUE
+cr_set_hairline (VALUE self, VALUE hairline)
+{
+  cairo_set_hairline (_SELF, RTEST (hairline));
+  cr_check_status (_SELF);
+  return self;
+}
+#endif
+
 static VALUE
 cr_set_line_cap (VALUE self, VALUE cap)
 {
@@ -567,7 +591,7 @@ cr_set_dash (int argc, VALUE *argv, VALUE self)
     {
       int i, length;
       double *values;
-      length = RARRAY_LEN (dash_array);
+      length = (int) RARRAY_LEN (dash_array);
       values = ALLOCA_N (double, length);
       if (!values)
         {
@@ -1323,7 +1347,7 @@ cr_show_text_glyphs (VALUE self, VALUE rb_utf8, VALUE rb_glyphs,
 
   cr = _SELF;
   utf8 = RSTRING_PTR (rb_utf8);
-  utf8_len = RSTRING_LEN (rb_utf8);
+  utf8_len = (int) RSTRING_LEN (rb_utf8);
   rb_cairo__glyphs_from_ruby_object (rb_glyphs, &glyphs, &num_glyphs);
   rb_cairo__text_clusters_from_ruby_object (rb_clusters,
                                             &clusters, &num_clusters);
@@ -1484,6 +1508,14 @@ cr_get_line_width (VALUE self)
   return rb_float_new (cairo_get_line_width (_SELF));
 }
 
+#if CAIRO_CHECK_VERSION(1, 17, 6)
+static VALUE
+cr_get_hairline (VALUE self)
+{
+  return CBOOL2RVAL (cairo_get_hairline (_SELF));
+}
+#endif
+
 static VALUE
 cr_get_line_cap (VALUE self)
 {
@@ -1571,20 +1603,26 @@ static VALUE
 cr_copy_path (VALUE self)
 {
   cairo_path_t *path;
+  VALUE rb_path;
 
   path = cairo_copy_path (_SELF);
   rb_cairo_check_status (path->status);
-  return CRPATH2RVAL (path);
+  rb_path = CRPATH2RVAL (path);
+  cairo_path_destroy (path);
+  return rb_path;
 }
 
 static VALUE
 cr_copy_path_flat (VALUE self)
 {
   cairo_path_t *path;
+  VALUE rb_path;
 
-  path =  cairo_copy_path_flat (_SELF);
+  path = cairo_copy_path_flat (_SELF);
   rb_cairo_check_status (path->status);
-  return CRPATH2RVAL (path);
+  rb_path = CRPATH2RVAL (path);
+  cairo_path_destroy (path);
+  return rb_path;
 }
 
 static VALUE
@@ -1745,6 +1783,9 @@ Init_cairo_context (void)
   rb_define_method (rb_cCairo_Context, "set_antialias", cr_set_antialias, 1);
   rb_define_method (rb_cCairo_Context, "set_fill_rule", cr_set_fill_rule, 1);
   rb_define_method (rb_cCairo_Context, "set_line_width", cr_set_line_width, 1);
+#if CAIRO_CHECK_VERSION(1, 17, 6)
+  rb_define_method (rb_cCairo_Context, "set_hairline", cr_set_hairline, 1);
+#endif
   rb_define_method (rb_cCairo_Context, "set_line_cap", cr_set_line_cap, 1);
   rb_define_method (rb_cCairo_Context, "set_line_join", cr_set_line_join, 1);
   rb_define_method (rb_cCairo_Context, "set_dash", cr_set_dash, -1);
@@ -1856,6 +1897,9 @@ Init_cairo_context (void)
                     cr_get_current_point, 0);
   rb_define_method (rb_cCairo_Context, "fill_rule", cr_get_fill_rule, 0);
   rb_define_method (rb_cCairo_Context, "line_width", cr_get_line_width, 0);
+#if CAIRO_CHECK_VERSION(1, 17, 6)
+  rb_define_method (rb_cCairo_Context, "hairline?", cr_get_hairline, 0);
+#endif
   rb_define_method (rb_cCairo_Context, "line_cap", cr_get_line_cap, 0);
   rb_define_method (rb_cCairo_Context, "line_join", cr_get_line_join, 0);
   rb_define_method (rb_cCairo_Context, "miter_limit", cr_get_miter_limit, 0);
@@ -1874,6 +1918,8 @@ Init_cairo_context (void)
 
   rb_define_method (rb_cCairo_Context, "to_ptr", cr_to_ptr, 0);
 
+  rb_define_method (rb_cCairo_Context, "raw_address", cr_raw_address, 0);
+
 #if CAIRO_CHECK_VERSION(1, 15, 4)
   /* Logical structure tagging functions */
   {
